package de.brightbyte.application;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.util.Arrays;
import java.util.Collections;
import java.util.InputMismatchException;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.brightbyte.io.Output;
import de.brightbyte.io.WriterOutput;
import de.brightbyte.util.StringUtils;
import de.brightbyte.util.StructuredDataCodec;

public abstract class ConsoleApp {
	
	public interface Command {
		public void execute(ConsoleApp app) throws Exception;
	}
	
	public static abstract class AbstractSimpleCommand implements Command {

		protected String command;
		protected List<Object> args;
		
		public AbstractSimpleCommand(String command, List<Object> args) {
			super();
			this.command = command;
			this.args = args==null ? Collections.<Object>emptyList() : args;
		}

		public abstract void execute(ConsoleApp app) throws Exception;

		@Override
		public String toString() {
			return command + " " + args;
		}

		@Override
		public int hashCode() {
			final int PRIME = 31;
			int result = 1;
			result = PRIME * result + ((args == null) ? 0 : args.hashCode());
			result = PRIME * result + ((command == null) ? 0 : command.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			final AbstractSimpleCommand other = (AbstractSimpleCommand) obj;
			if (args == null) {
				if (other.args != null)
					return false;
			} else if (!args.equals(other.args))
				return false;
			if (command == null) {
				if (other.command != null)
					return false;
			} else if (!command.equals(other.command))
				return false;
			return true;
		}
		
		
	}
	
	public static class Echo extends AbstractSimpleCommand {

		public Echo(List<Object> args) {
			super("echo", args);
		}

		@Override
		public void execute(ConsoleApp app) throws Exception {
			app.echo(StringUtils.join(" ", args));
		}
		
	}
	
	public static class Quit extends AbstractSimpleCommand {

		public Quit() {
			super("quit", null);
		}

		@Override
		public void execute(ConsoleApp app) throws Exception {
			app.echo("bye");
			app.quit = true;
		}
		
	}
	
	
	protected BufferedReader in;
	protected Output out;
	protected boolean quit;
	
	protected Pattern comment = Pattern.compile("([#;%]|--|//).*");
	protected Pattern commandMatcher = Pattern.compile("^\\w[\\w\\d]*");
	protected Pattern argumentSplitter = Pattern.compile("\\s+");
	private String prompt = "> ";
	
	public ConsoleApp() {
		this(System.in, System.out);
	}
	
	public ConsoleApp(InputStream in, OutputStream out) {
		this( new InputStreamReader(in),
				new OutputStreamWriter(out));
	}
	
	public ConsoleApp(InputStream in, OutputStream out, String enc) throws UnsupportedEncodingException {
		this( new InputStreamReader(in, enc),
				new OutputStreamWriter(out, enc) );
	}
	
	public ConsoleApp(Reader in, Writer out) {
		this( new BufferedReader(in), new WriterOutput(out) );
	}
	
	public ConsoleApp(BufferedReader in, Output out) {
		super();
		this.in = in;
		this.out = out;
	}
	
	public String readLine() throws IOException {
		return in.readLine();
	}

	public void run() {
		while (!quit) {
			try {
				process();
			}
			catch (Exception ex) {
				echoError("command failed", ex);
			}
		}
	}

	protected String prompt() throws IOException {
		out.write(prompt );
		out.flush();
		String line = readLine();
		return line;
	}
	
	protected void process() throws Exception {
		String line;
		
		try { 
			line = prompt();
		}
		catch (IOException ex) {
			echoError("input failed", ex);
			quit = true;
			return;
		}
			
		if (line==null) {
			quit = true;
			return;
		}
		
		line = line.trim();

		if (line.length()==0) {
			return;
		}

		if (comment.matcher(line).matches()) {
			return;
		}
		
		Command cmd;
		try {
			cmd = readCommand(line);
			if (cmd==null) return;
		}
		catch (InputMismatchException ex) {
			echoError("bad input", ex);
			return;
		}
		catch (IOException ex) {
			echoError("input failed", ex);
			quit = true;
			return;
		}
		
		cmd.execute(this);
	}

	protected Command readCommand(String line) throws IOException {
		Matcher m = commandMatcher.matcher(line);
		if (!m.find()) {
			echo("bad input: %s", line);
			return null;
		}
		
		String c = m.groupCount() > 0 ? m.group(1) : m.group();
		
		List<Object> args = getArguments(m.end(), line);
		
		Command cmd = newCommand(c, args);
		
		if (cmd==null) {
			echo("bad command: %s", line);
		}
		
		return cmd;
	}

	@SuppressWarnings("unchecked")
	private List<Object> getArguments(int off, String line) {
		String a = line.substring(off).trim();
		if (a.startsWith("[") || a.startsWith("(")) {
			return (List<Object>)StructuredDataCodec.instance.decodeValue(a);
		}
		else {
			Object[] aa = argumentSplitter.split(a);
			return Arrays.asList(aa);
		}
	}

	protected Command newCommand(String cmd, List<Object> args) {
		if (cmd.equals("echo")) return new Echo(args);
		else if (cmd.equals("quit")) return new Quit();
		else return null;
	}

	public void echoError(String msg, Exception ex) {
		echo(msg + " " + ex);
	}

	public void echo(String msg, Object... args) {
		try {
			if (args.length>0) msg = String.format(msg, args);
			out.writeln(msg);
		} catch (IOException e) {
			System.err.println(e);
			quit = true;
		}
	}
	
	public static void main(String[] args) throws UnsupportedEncodingException {
		ConsoleApp app = new ConsoleApp() {
			//..
		};
		
		app.run();
	}

}
