package de.brightbyte.data;

import java.io.IOException;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.util.Arrays;

import de.brightbyte.util.StringUtils;

public class ByteString implements Comparable<ByteString>, Serializable {
	private static final long serialVersionUID = 6148088195249530279L; //NOTE: refer to custom readObject and writeObject methods.

	public static final ByteString EMPTY = new ByteString(new byte[0]);
	
	private byte[] data;
	private int offset;
	private int length;
	
	private transient boolean hashed = false;
	private transient int hash;

	public ByteString(String s, String charset) throws UnsupportedEncodingException {
		this(s.getBytes(charset));
	}
	
	public ByteString(byte[] data) {
		this(data, 0, data.length, true);
	}
	
	public ByteString(byte[] data, int offset, int length) {
		this(data, offset, length, true);
	}
	
	public ByteString(ByteString s, int offset, int length) {
		this(s.data, s.offset+offset, length, true);
	}
	
	private ByteString(byte[] data, int offset, int length, boolean clone) {
		    if (offset+length>data.length) throw new IndexOutOfBoundsException("length exceeds available data: "+offset+length+" > "+data.length);
		
			if (clone) {
				byte[] buff = new byte[length];
				System.arraycopy(data, offset, buff, 0, length);
				this.data = buff;
				this.offset = 0;
				this.length = length;
			} else {
				this.data = data;
				this.offset = offset;
				this.length = length;
			}
	}
	
	public void getBytes(int srcBegin, int srcEnd, byte[] dst, int dstBegin)  {
		System.arraycopy(data, srcBegin+offset, dst, dstBegin, srcEnd-srcBegin);
	}
	
	public byte byteAt(int index) {
		if (index>=length) throw new IndexOutOfBoundsException("out of bounds: "+index);
		if (index<0) throw new IndexOutOfBoundsException("negative index: "+index);
		return data[offset+index];
	}

	public int length() {
		return length;
	}

	public ByteString subBString(int start, int end) {
		return new ByteString(data, offset+start, end-start);
	}
	
	public String toString() {
		return StringUtils.hex(data, offset, length);
	}

	@Override
	public int hashCode() {
		if (hashed) return hash;
		
		final int PRIME = 31;
		int hash = 1;
		hash = PRIME * hash + Arrays.hashCode(data);
		hash = PRIME * hash + length;
		hash = PRIME * hash + offset;
		
		hashed = true;
		return hash;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (! (obj instanceof ByteString))
			return false;
		
		final ByteString other = (ByteString) obj;

		if (other.length!=this.length) 
			return false;
		if (other.data== this.data) 
			return true;
		if (other.hashCode() != this.hashCode()) 
			return false;

		for (int i=0; i<length; i++) {
			if (this.data[this.offset+i]!=other.data[other.offset]) 
				return false;
		}

		return true;
	}

	public int compareTo(ByteString other) {
		for (int i=0; i<length && i<other.length; i++) {
			if (this.data[this.offset+i]<other.data[other.offset]) return -1; 
			else if (this.data[this.offset+i]>other.data[other.offset]) return +1; 
		}
		
		if (other.length>this.length) return -1;
		else if (other.length<this.length) return +1;

		return 0;
	}

	 @SuppressWarnings("unused")
	 private void writeObject(java.io.ObjectOutputStream out)
	 	throws IOException {
		 	out.writeInt(length);
		 	out.write(data, offset, length);
	 }
	 
	 @SuppressWarnings("unused")
	 private void readObject(java.io.ObjectInputStream in)
     	throws IOException {
		 	length = in.readInt();
		 	offset = 0;
		 	data = new byte[length];
	 		in.read(data, offset, length);
	 }
	 
	 @SuppressWarnings("unused")
	private void readObjectNoData() {
		 data = new byte[0];
		 offset = 0;
		 length = 0;
	 }
	
}
