package de.brightbyte.data;

import java.util.AbstractMap;
import java.util.AbstractSet;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.zip.DataFormatException;

import de.brightbyte.io.BlockCodec;

public class CompressedKeyMap<K, V> implements Map<K, V> {
	protected class EntryIterator implements Iterator<Map.Entry<K, V>> {
		protected Iterator<Map.Entry<ByteString, V>> iterator = map.entrySet().iterator();

		public boolean hasNext() {
			return iterator.hasNext();
		}

		public Map.Entry<K, V> next() {
			Map.Entry<ByteString, V> e = iterator.next();
			K k = uncompressKey(e.getKey());
			
			return new AbstractMap.SimpleEntry<K, V>(k, e.getValue());
		}

		public void remove() {
			iterator.remove();
		}
	}
	
	protected class KeyIterator implements Iterator<K> {
		protected Iterator<ByteString> iterator = map.keySet().iterator();

		public boolean hasNext() {
			return iterator.hasNext();
		}

		public K next() {
			ByteString b = iterator.next();
			K k = uncompressKey(b);
			return k;
		}

		public void remove() {
			iterator.remove();
		}
	}
	
		protected class EntrySet extends AbstractSet<Map.Entry<K, V>> {
			@Override
			public Iterator<java.util.Map.Entry<K, V>> iterator() {
				return new EntryIterator();
			}

			@Override
			public int size() {
				return CompressedKeyMap.this.size();
			}

			@Override
			public void clear() {
				CompressedKeyMap.this.clear();
			}

			@Override
			public boolean add(java.util.Map.Entry<K, V> e) {
				V v = e.getValue();
				V old = put(e.getKey(), e.getValue());
				
				if (v==old) return false;
				else if (v==null || old==null) return true;
				else return !v.equals(old);
			}
		}
	
		protected class KeySet extends AbstractSet<K> {
			@Override
			public Iterator<K> iterator() {
				return new KeyIterator();
			}

			@Override
			public int size() {
				return CompressedKeyMap.this.size();
			}

			@Override
			public void clear() {
				CompressedKeyMap.this.clear();
			}

			@Override
			public boolean contains(Object o) {
				return CompressedKeyMap.this.containsKey(o);
			}

			@Override
			public boolean remove(Object o) {
				V old = CompressedKeyMap.this.remove(o);
				return old!=null;
			}
		}
	
		protected Map<ByteString, V> map;
		protected BlockCodec<K> codec;
		
		private int hash = 0;
		
		private Set<K> theKeySet = null;
		private Set<Map.Entry<K, V>> theEntrySet = null;

		public void clear() {
			map.clear();
		}
		
		@SuppressWarnings("unchecked")
		protected ByteString compressKey(Object key) {
			return codec.encode((K)key);
		}
		
		@SuppressWarnings("unchecked")
		protected K uncompressKey(ByteString k) {
			try {
				return codec.decode(k);
			} catch (DataFormatException e) {
				throw new RuntimeException("codec failed to decode compressed key: " + k, e);
			}
		}
		
		public boolean containsKey(Object key) {
			return map.containsKey(compressKey(key));
		}

		public boolean containsValue(Object value) {
			return map.containsValue(value);
		}

		public Set<Entry<K, V>> entrySet() {
			if (theEntrySet==null) theEntrySet = new EntrySet();
			return theEntrySet;
		}

		public boolean equals(Object o) {
			if (o==this) return true;
			if (o==null) return false;
			if (!(o instanceof Map)) return false;
			
			Map m = (Map)o;
			if (m.size() != this.size()) return false; 
			if (m.hashCode() != this.hashCode()) return false; 
			
			for(Map.Entry<? extends K, ? extends V> e: entrySet()) {
				Object ov =  m.get(e.getKey());
				Object v = e.getValue();
				
				if (ov!=v) {
					if (v==null || ov==null) return false;
					if (!v.equals(ov)) return false;
				}
			}

			return true;
		}

		public V get(Object key) {
			return map.get(compressKey(key));
		}

		public int hashCode() {
			return hash;
		}

		public boolean isEmpty() {
			return map.isEmpty();
		}

		public Set<K> keySet() {
			if (theKeySet==null) theKeySet = new KeySet();
			return theKeySet;
		}

		public V put(K key, V value) {
			hash ^= key.hashCode();
			if (value!=null) hash ^= value.hashCode();
			
			V v = map.put(compressKey(key), value);

			if (v!=null) hash ^= v.hashCode();

			return v;
		}

		public void putAll(Map<? extends K, ? extends V> m) {
			for(Map.Entry<? extends K, ? extends V> e: m.entrySet()) {
				put(e.getKey(), e.getValue());
			}
		}

		public V remove(Object key) {
			V v = map.remove(compressKey(key));

			if (v!=null) {
				hash ^= key.hashCode();
				hash ^= v.hashCode();
			}
			
			return v;
		}

		public int size() {
			return map.size();
		}

		public Collection<V> values() {
			return map.values();
		}
		
}
