package de.brightbyte.data;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

public class Functors {
	public static <S, T>List<T> apply(List<S> data, Functor<T, S> f) {
		ArrayList<T> a = new ArrayList<T>(data.size());
		for (S s : data) {
			T t = f.apply(s);
			a.add(t);
		}
		return a;
	}
	
	public static <K, S, T>Map<K, T> apply(Map<K, S> data, Functor<T, S> f) {
		HashMap<K, T> a = new HashMap<K, T>(data.size());
		for (Map.Entry<K, S> e : data.entrySet()) {
			T t = f.apply(e.getValue());
			a.put(e.getKey(), t);
		}
		return a;
	}

	public static <T> T fold(Collection<T> data, Functor2<T, T, T> f) {
		T akk = null;
		for (T t : data) {
			akk = f.apply(akk, t);
		}
		
		return akk;
	}
	
	public static <T>Functor<T, ? extends Collection<T>> aggregator(final Functor2<? extends T, T, T> aggregat) {
		return new Functor<T, Collection<T>>() {
			public T apply(Collection<T> list) {
				Iterator<T> it = list.iterator();
				if (!it.hasNext()) return null;
				
				T v = it.next();
				while (it.hasNext()) {
					v = aggregat.apply(v, it.next());
				}
				return v;
			}
		};
	}
	
	public static <T>Functor<T, ? extends Collection<T>> firstElement() {
		return new Functor<T, Collection<T>>() {
			public T apply(Collection<T> list) {
				Iterator<T> it = list.iterator();
				if (!it.hasNext()) return null;
				else return it.next();
			}
		};
	}
	
	public static <T extends Number>Functor<T, ? extends Collection<T>> max(Comparator<T> comparator) {
		return aggregator(max2(comparator));
	}
	
	public static <T extends Number>Functor2<T, T, T> max2(Comparator<T> comparator) {
		if (comparator==null) comparator = (Comparator<T>)(Object)NaturalComparator.instance;
		final Comparator<T> comp = comparator;
		
		return new Functor2<T, T, T>() {
			public T apply(T b, T c) {
				return comp.compare(b, c)>0 ? b : c;
			}
		};
	}

	public static <T extends Collection<String>>Functor<String, ? extends T> concat() {
		return new Functor<String, T>() {
			public String apply(T b) {
				StringBuilder s = new StringBuilder();
				
				for (String x: b) s.append(x);
				return s.toString();
			}
		};
	}
	
	public static <T extends Collection<? extends Object>>Functor<String, ? extends T> concat(final String separator) {
		return new Functor<String, T>() {
			public String apply(T b) {
				StringBuilder s = new StringBuilder();
				
				boolean first = true;
				for (Object x: b) {
					if (first) first = false;
					else s.append(separator);
					
					s.append(x);
				}
				
				return s.toString();
			}
		};
	}
	
	public static class Double {
		public static final Functor2<java.lang.Double, Number, Number> max2 = new Functor2<java.lang.Double, Number, Number>() {
			public java.lang.Double apply(Number b, Number c) {
				return b.doubleValue() > c.doubleValue() ? b.doubleValue() : c.doubleValue();
			}
		};
		
		public static final Functor2<java.lang.Double, Number, Number> sum2 = new Functor2<java.lang.Double, Number, Number>() {
			public java.lang.Double apply(Number b, Number c) {
				return b.doubleValue() + c.doubleValue();
			}
		};
		
		public static final Functor2<java.lang.Double, Number, Number> product2 = new Functor2<java.lang.Double, Number, Number>() {
			public java.lang.Double apply(Number b, Number c) {
				return b.doubleValue() * c.doubleValue();
			}
		};
		
		public static final Functor<Number, ? extends Collection<? extends Number>> sum = Functors.aggregator(sum2);
		public static final Functor<Number, ? extends Collection<? extends Number>> product = Functors.aggregator(product2);
		public static final Functor<Number, ? extends Collection<? extends Number>> max = Functors.aggregator(max2);
		
		public static final Comparator<Number> comparator = new Comparator<Number>(){
			public int compare(Number a, Number b) {
				double x = a.doubleValue();
				double y = b.doubleValue();
				
				if (x==y ) return 0;
				else if (x>y) return 1;
				else return -1;
			}
		};
	}
	
	public static class Integer {
		public static final Functor2<java.lang.Integer, Number, Number> max2 = new Functor2<java.lang.Integer, Number, Number>() {
			public java.lang.Integer apply(Number b, Number c) {
				return b.intValue() > c.intValue() ? b.intValue() : c.intValue();
			}
		};
		
		public static final Functor2<java.lang.Integer, Number, Number> sum2 = new Functor2<java.lang.Integer, Number, Number>() {
			public java.lang.Integer apply(Number b, Number c) {
				return b.intValue() + c.intValue();
			}
		};
		
		public static final Functor2<java.lang.Integer, Number, Number> product2 = new Functor2<java.lang.Integer, Number, Number>() {
			public java.lang.Integer apply(Number b, Number c) {
				return b.intValue() * c.intValue();
			}
		};
		
		public static final Functor<Number, ? extends Collection<? extends Number>> sum = Functors.aggregator(sum2);
		public static final Functor<Number, ? extends Collection<? extends Number>> product = Functors.aggregator(product2);
		public static final Functor<Number, ? extends Collection<? extends Number>> max = Functors.aggregator(max2);
		
		public static final Comparator<Number> comparator = new Comparator<Number>(){
			public int compare(Number a, Number b) {
				return a.intValue() - b.intValue();
			}
		};
	}
	
	public static class Long {
		public static final Functor2<java.lang.Long, Number, Number> max2 = new Functor2<java.lang.Long, Number, Number>() {
			public java.lang.Long apply(Number b, Number c) {
				return b.intValue() > c.intValue() ? b.longValue() : c.longValue();
			}
		};

		public static final Functor2<java.lang.Long, Number, Number> sum2 = new Functor2<java.lang.Long, Number, Number>() {
			public java.lang.Long apply(Number b, Number c) {
				return b.longValue() + c.longValue();
			}
		};
		
		public static final Functor2<java.lang.Long, Number, Number> product2 = new Functor2<java.lang.Long, Number, Number>() {
			public java.lang.Long apply(Number b, Number c) {
				return b.longValue() * c.longValue();
			}
		};
		
		public static final Functor<Number, ? extends Collection<? extends Number>> sum = Functors.aggregator(sum2);
		public static final Functor<Number, ? extends Collection<? extends Number>> product = Functors.aggregator(product2);
		public static final Functor<Number, ? extends Collection<? extends Number>> max = Functors.aggregator(max2);
		
		public static final Comparator<Number> comparator = new Comparator<Number>(){
			public int compare(Number a, Number b) {
				long x = a.longValue();
				long y = b.longValue();
				
				if (x==y ) return 0;
				else if (x>y) return 1;
				else return -1;
			}
		};
	}
	
	public static class BigDecimal {
		public static final Functor2<java.math.BigDecimal, java.math.BigDecimal, java.math.BigDecimal> max2 = new Functor2<java.math.BigDecimal, java.math.BigDecimal, java.math.BigDecimal>() {
			public java.math.BigDecimal apply(java.math.BigDecimal b, java.math.BigDecimal c) {
				return b.compareTo(c) > 0 ? b : c;
			}
		};

		public static final Functor2<java.math.BigDecimal, java.math.BigDecimal, java.math.BigDecimal> sum2 = new Functor2<java.math.BigDecimal, java.math.BigDecimal, java.math.BigDecimal>() {
			public java.math.BigDecimal apply(java.math.BigDecimal b, java.math.BigDecimal c) {
				return b.add(c);
			}
		};
		
		public static final Functor2<java.math.BigDecimal, java.math.BigDecimal, java.math.BigDecimal> product2 = new Functor2<java.math.BigDecimal, java.math.BigDecimal, java.math.BigDecimal>() {
			public java.math.BigDecimal apply(java.math.BigDecimal b, java.math.BigDecimal c) {
				return b.multiply(c);
			}
		};
		
		public static final Functor<java.math.BigDecimal, ? extends Collection<? extends java.math.BigDecimal>> sum = Functors.aggregator(sum2);
		public static final Functor<java.math.BigDecimal, ? extends Collection<? extends java.math.BigDecimal>> product = Functors.aggregator(product2);
		public static final Functor<java.math.BigDecimal, ? extends Collection<? extends java.math.BigDecimal>> max = Functors.aggregator(max2);
		
		public static final Comparator<java.math.BigDecimal> comparator = new Comparator<java.math.BigDecimal>(){
			public int compare(java.math.BigDecimal a, java.math.BigDecimal b) {
				return a.compareTo(b);
			}
		};
	}
	
	public static class BigInteger {
		public static final Functor2<java.math.BigInteger, java.math.BigInteger, java.math.BigInteger> max2 = new Functor2<java.math.BigInteger, java.math.BigInteger, java.math.BigInteger>() {
			public java.math.BigInteger apply(java.math.BigInteger b, java.math.BigInteger c) {
				return b.compareTo(c) > 0 ? b : c;
			}
		};

		public static final Functor2<java.math.BigInteger, java.math.BigInteger, java.math.BigInteger> sum2 = new Functor2<java.math.BigInteger, java.math.BigInteger, java.math.BigInteger>() {
			public java.math.BigInteger apply(java.math.BigInteger b, java.math.BigInteger c) {
				return b.add(c);
			}
		};
		
		public static final Functor2<java.math.BigInteger, java.math.BigInteger, java.math.BigInteger> product2 = new Functor2<java.math.BigInteger, java.math.BigInteger, java.math.BigInteger>() {
			public java.math.BigInteger apply(java.math.BigInteger b, java.math.BigInteger c) {
				return b.multiply(c);
			}
		};
		
		public static final Comparator<java.math.BigInteger> comparator = new Comparator<java.math.BigInteger>(){
			public int compare(java.math.BigInteger a, java.math.BigInteger b) {
				return a.compareTo(b);
			}
		};
		
		public static final Functor<java.math.BigInteger, ? extends Collection<? extends java.math.BigInteger>> sum = Functors.aggregator(sum2);
		public static final Functor<java.math.BigInteger, ? extends Collection<? extends java.math.BigInteger>> product = Functors.aggregator(product2);
		public static final Functor<java.math.BigInteger, ? extends Collection<? extends java.math.BigInteger>> max = Functors.aggregator(max2);
	}
}
