package de.brightbyte.text;

import java.text.ParseException;
import java.text.ParsePosition;

public class DefaultStringLiteralParser implements StringLiteralParser {
	public static final int ALLOW_NAMED_ESCAPE = 1;
	public static final int ALLOW_HEX_ESCAPE = 2;
	public static final int ALLOW_OCT_ESCAPE = 4;
	public static final int ALLOW_UNICODE_ESCAPE = 8;
	public static final int ALLOW_REPEATED_QUOTE = 16;
	public static final int ALLOW_LINE_BREAKS = 32;
	public static final int ALLOW_ESCAPED_LINE_BREAKS = 64;
	public static final int IMPLICIT_INITIAL_QUOTE = 256;
	public static final int IMPLICIT_FINAL_QUOTE = 512;
	
	public static final DefaultStringLiteralParser instance = new DefaultStringLiteralParser();
	
	protected String quotes;
	protected char escape;
	protected int flags;
	
	public DefaultStringLiteralParser() {
		this("\"'");
	}
	
	public DefaultStringLiteralParser(String quotes) {
		this(quotes, '\\', ALLOW_NAMED_ESCAPE | ALLOW_HEX_ESCAPE | ALLOW_UNICODE_ESCAPE );
	}
	
	public DefaultStringLiteralParser(String quotes, char escape, int flags) {
		this.quotes = quotes;
		this.escape = escape;
		this.flags = flags;
	}
	
	public String parseStringLiteral(String s, ParsePosition pos) throws ParseException {
		int i = pos.getIndex();
		StringBuilder b = new StringBuilder();

		int escapeStart = 0;
		int escapeMode = ' ';
		int escapeDigits = 0; 
		
		char quote = '\0';		
		int state = 0;
		
		if ((flags & IMPLICIT_INITIAL_QUOTE)>0) {
			quote = quotes.charAt(0);
			state = 1;
		}
		
		while (state < 1000) {
			int ch = i<s.length() ? s.charAt(i) : -1;
			
			switch (state) {
				case 0: //expect initial quote
					if (quotes.indexOf((char)ch)>=0) {
						quote = (char)ch;
						state = 1;
					}
					else throw new ParseException("expected "+quote, i);
					break;
					
				case 1: //string body
					if (ch==-1) {
						if ((flags & IMPLICIT_FINAL_QUOTE)>0) state = 1000; 
						else throw new ParseException("unexpected end of input, expected "+quote, i);
					}
					else if (ch==quote) {
						if ((flags & ALLOW_REPEATED_QUOTE)>0) state = 3; //possible end
						else state = 1000; //definite end
					}
					else if (escape!='\0' && ch==escape) state = 2; //escaped
					else if ((flags & ALLOW_LINE_BREAKS)==0 && (ch=='\r' || ch=='\n')) {
						if ((flags & IMPLICIT_FINAL_QUOTE)>0) {
							state = 1000;
							i--;
						}
						else throw new ParseException("unexpected end of line, expected "+quote, i);
					}
					else b.append((char)ch);
					break;

				case 2: //escaped
					state = 1;
					if (ch==-1) throw new ParseException("unexpected end of input after "+escape, i);
					else if (ch=='t' && (flags & ALLOW_NAMED_ESCAPE)>0) b.append('\t'); 
					else if (ch=='b' && (flags & ALLOW_NAMED_ESCAPE)>0) b.append('\b'); 
					else if (ch=='n' && (flags & ALLOW_NAMED_ESCAPE)>0) b.append('\n'); 
					else if (ch=='f' && (flags & ALLOW_NAMED_ESCAPE)>0) b.append('\f'); 
					else if (ch=='r' && (flags & ALLOW_NAMED_ESCAPE)>0) b.append('\r'); 
					else if (ch=='0' && (flags & ALLOW_OCT_ESCAPE)>0) {
						state = 10;
						escapeStart = i+1;
						escapeMode = '0';
						escapeDigits = 3;
					}
					else if (ch=='x' && (flags & ALLOW_HEX_ESCAPE)>0) {
						state = 10;
						escapeStart = i+1;
						escapeMode = 'x';
						escapeDigits = 2;
					}
					else if (ch=='u' && (flags & ALLOW_UNICODE_ESCAPE)>0) {
						state = 10;
						escapeStart = i+1;
						escapeMode = 'x';
						escapeDigits = 4;
					}
					else if (ch=='\r') {
						if ((flags & ALLOW_ESCAPED_LINE_BREAKS)>0) {
							b.append('\r');
							state = 6;
						}
						else throw new ParseException("escaped line breaks are not supported", i);
					}
					else if (ch=='\n') {
						if ((flags & ALLOW_ESCAPED_LINE_BREAKS)>0) b.append((char)ch);
						else throw new ParseException("escaped line breaks are not supported", i);
					}
					else if (ch=='0' && (flags & ALLOW_NAMED_ESCAPE)>0) b.append('\0'); 
					else b.append((char)ch);
					break;
					
				case 3: //end or repeated quote 
					if (ch==quote) {
						b.append(quote);
						state = 1;
					} else {
						i--; //pushback
						state = 1000; //terminate
					}
					break;

				case 6: //escaped line break 
					if (ch==-1) {
						if ((flags & IMPLICIT_FINAL_QUOTE)>0) {
							state = 1000; 
						}
						else throw new ParseException("unexpected end of input, expected "+quote, i);
					}
					else if (ch=='\n') {
						b.append('\n');
						state = 1;
					} else {
						i--;
						state = 1;
					}
					break;

				case 10: //end or repeated quote
					if (ch==-1) throw new ParseException("unexpected end of input inside escape sequence", i);
					else if (escapeMode=='0' && ch>='0' && ch<='7') ;
					else if (escapeMode=='x' && ((ch>='0' && ch<='9') || (ch>='A' && ch<='F') || (ch>='a' && ch<='f'))) ;
					else if (escapeMode=='u' && ((ch>='0' && ch<='9') || (ch>='A' && ch<='F') || (ch>='a' && ch<='f'))) ;
					else throw new ParseException("unexpected character in escape sequence", i);

					escapeDigits--;
					if (escapeDigits<=0) {
						int c;
						
						if (escapeMode=='0') c = Integer.parseInt(s.substring(escapeStart, i+1), 8);
						else if (escapeMode=='x') c = Integer.parseInt(s.substring(escapeStart, i+1), 16);
						else if (escapeMode=='u') c = Integer.parseInt(s.substring(escapeStart, i+1), 16);
						else throw new IllegalArgumentException("bad escape mode");
						
						b.append((char)c);
						state = 1;
					}
					break;
					
			}
			
			i++;
		}

		pos.setIndex(i);
		return b.toString();
	}

	public String getQuoteChars() {
		return quotes;
	}
	
	public static String scanTo(String s, ParsePosition pos, String find, boolean not) {
		int start = pos.getIndex();
		int i = start; 
		while (i<s.length()) {
			char ch = s.charAt(i);
			if ( (find.indexOf(ch)<0) == not) break;
			i++;
		}
		
		pos.setIndex(i);
		return s.substring(start, i);
	}
}
