package de.brightbyte.text;

import java.text.ParseException;
import java.text.ParsePosition;

import junit.framework.TestCase;

import static de.brightbyte.text.DefaultStringLiteralParser.*;

public class DefaultStringLiteralParserTest extends TestCase {

	public void testScanTo() {
		ParsePosition pos = new ParsePosition(1);
		
		assertEquals("abc", DefaultStringLiteralParser.scanTo("_abcxyz", pos, "xyz", false));
		assertEquals(4, pos.getIndex());

		pos.setIndex(1);
		assertEquals("abc", DefaultStringLiteralParser.scanTo("_abcxyz", pos, "abc", true));
		assertEquals(4, pos.getIndex());
	}

	public void testParseStringLiteral() throws ParseException {
		DefaultStringLiteralParser parser = new DefaultStringLiteralParser("\"'", '\\', 0);
		ParsePosition pos = new ParsePosition(0);
		
		assertEquals("abc", parser.parseStringLiteral("\"abc\"\"efg\"", pos));
		assertEquals(5, pos.getIndex());
		
		assertEquals("efg", parser.parseStringLiteral("\"abc\"\"efg\"", pos));
		assertEquals(10, pos.getIndex());

		assertEquals("a'c", parser.parseStringLiteral("\"a'c\"", new ParsePosition(0)));
		assertEquals("a\"c", parser.parseStringLiteral("'a\"c'", new ParsePosition(0)));
		assertEquals("a\"c", parser.parseStringLiteral("\"a\\\"c\"", new ParsePosition(0)));
		assertEquals("a\\c", parser.parseStringLiteral("\"a\\\\c\"", new ParsePosition(0)));

		assertEquals("", parser.parseStringLiteral("''", new ParsePosition(0)));

		try {
			parser.parseStringLiteral("\"abc'", new ParsePosition(0));
			fail("expected complaint about unexpected end of input");
		} catch (ParseException ex) {
			//pass
		}
		
		parser = new DefaultStringLiteralParser("'", '\0', 0);
		pos.setIndex(0);
		assertEquals("a\\", parser.parseStringLiteral("'a\\'", pos));
	}

	public void testParseStringLiteral_ALLOW_NAMED_ESCAPE() throws ParseException {
		DefaultStringLiteralParser parser = new DefaultStringLiteralParser("'", '\\', 0);
		assertEquals("anb", parser.parseStringLiteral("'a\\nb'", new ParsePosition(0)));
		
		parser = new DefaultStringLiteralParser("'", '\\', ALLOW_NAMED_ESCAPE);
		assertEquals("a\rb", parser.parseStringLiteral("'a\\rb'", new ParsePosition(0)));
		assertEquals("a\nb", parser.parseStringLiteral("'a\\nb'", new ParsePosition(0)));
		assertEquals("a\tb", parser.parseStringLiteral("'a\\tb'", new ParsePosition(0)));
		assertEquals("a\bb", parser.parseStringLiteral("'a\\bb'", new ParsePosition(0)));
	}
	
	public void testParseStringLiteral_ALLOW_HEX_ESCAPE() throws ParseException {
		DefaultStringLiteralParser parser = new DefaultStringLiteralParser("'", '\\', 0);
		assertEquals("ax4Ab", parser.parseStringLiteral("'a\\x4Ab'", new ParsePosition(0)));
		
		parser = new DefaultStringLiteralParser("'", '\\', ALLOW_HEX_ESCAPE);
		assertEquals("aJb", parser.parseStringLiteral("'a\\x4Ab'", new ParsePosition(0)));
		
		try {
			parser.parseStringLiteral("'a\\x4", new ParsePosition(0));
			fail("expected complaint about unexpected end of escape sequence");
		} catch (ParseException ex) {
			//pass
		}
		
		try {
			parser.parseStringLiteral("'a\\x4X'", new ParsePosition(0));
			fail("expected complaint about bad escape sequence");
		} catch (ParseException ex) {
			//pass
		}
	}
	
	
	public void testParseStringLiteral_ALLOW_UNICODE_ESCAPE() throws ParseException {
		DefaultStringLiteralParser parser = new DefaultStringLiteralParser("'", '\\', 0);
		assertEquals("au00e4b", parser.parseStringLiteral("'a\\u00e4b'", new ParsePosition(0)));
		
		parser = new DefaultStringLiteralParser("'", '\\', ALLOW_UNICODE_ESCAPE);
		assertEquals("a\u00e4b", parser.parseStringLiteral("'a\\u00e4b'", new ParsePosition(0)));
		
		try {
			parser.parseStringLiteral("'a\\ue4", new ParsePosition(0));
			fail("expected complaint about unexpected end of escape sequence");
		} catch (ParseException ex) {
			//pass
		}
		
		try {
			parser.parseStringLiteral("'a\\u000xb'", new ParsePosition(0));
			fail("expected complaint about bad escape sequence");
		} catch (ParseException ex) {
			//pass
		}
	}
	
	public void testParseStringLiteral_ALLOW_OCT_ESCAPE() throws ParseException {
		DefaultStringLiteralParser parser = new DefaultStringLiteralParser("'", '\\', 0);
		assertEquals("a01123", parser.parseStringLiteral("'a\\01123'", new ParsePosition(0)));
		
		parser = new DefaultStringLiteralParser("'", '\\', ALLOW_OCT_ESCAPE);
		assertEquals("aJ3", parser.parseStringLiteral("'a\\01123'", new ParsePosition(0)));
		
		try {
			parser.parseStringLiteral("'a\\01", new ParsePosition(0));
			fail("expected complaint about unexpected end of escape sequence");
		} catch (ParseException ex) {
			//pass
		}
		
		try {
			parser.parseStringLiteral("'a\\0999'", new ParsePosition(0));
			fail("expected complaint about bad escape sequence");
		} catch (ParseException ex) {
			//pass
		}
	}

	public void testParseStringLiteral_ALLOW_REPEATED_QUOTE() throws ParseException {
		DefaultStringLiteralParser parser = new DefaultStringLiteralParser("'", '\\', 0);
		assertEquals("foo", parser.parseStringLiteral("'foo''bar'", new ParsePosition(0)));
		
		parser = new DefaultStringLiteralParser("'", '\\', ALLOW_REPEATED_QUOTE);
		assertEquals("foo'bar", parser.parseStringLiteral("'foo''bar'", new ParsePosition(0)));
		assertEquals("foo'", parser.parseStringLiteral("'foo'''", new ParsePosition(0)));
		
		try {
			parser.parseStringLiteral("'foo''", new ParsePosition(0));
			fail("expected complaint about unexpected end of input");
		} catch (ParseException ex) {
			//pass
		}
	}

	public void testParseStringLiteral_ALLOW_LINE_BREAKS() throws ParseException {
		DefaultStringLiteralParser parser = new DefaultStringLiteralParser("'", '\\', 0);

		try {
			parser.parseStringLiteral("foo\r\nbar", new ParsePosition(0));
			fail("expected complaint about unexpected end of line");
		} catch (ParseException ex) {
			//pass
		}
		
		parser = new DefaultStringLiteralParser("'", '\\', ALLOW_LINE_BREAKS);
		assertEquals("foo\r\nbar", parser.parseStringLiteral("'foo\r\nbar'", new ParsePosition(0)));
	}

	public void testParseStringLiteral_ALLOW_ESCAPED_BREAKS() throws ParseException {
		DefaultStringLiteralParser parser = new DefaultStringLiteralParser("'", '\\', 0);

		try {
			parser.parseStringLiteral("foo\\\r\nbar", new ParsePosition(0));
			fail("expected complaint about bad escape sequence");
		} catch (ParseException ex) {
			//pass
		}
		
		parser = new DefaultStringLiteralParser("'", '\\', ALLOW_ESCAPED_LINE_BREAKS);
		assertEquals("foo\rbar", parser.parseStringLiteral("'foo\\\rbar'", new ParsePosition(0)));
		assertEquals("foo\nbar", parser.parseStringLiteral("'foo\\\nbar'", new ParsePosition(0)));
		assertEquals("foo\r\nbar", parser.parseStringLiteral("'foo\\\r\nbar'", new ParsePosition(0)));
		assertEquals("foo\r", parser.parseStringLiteral("'foo\\\r'", new ParsePosition(0)));
	}

	public void testParseStringLiteral_IMPLICIT_INITIAL_QUOTE() throws ParseException {
		DefaultStringLiteralParser parser = new DefaultStringLiteralParser("'", '\\', 0);

		try {
			parser.parseStringLiteral("foo'", new ParsePosition(0));
			fail("expected complaint about missing initial quote");
		} catch (ParseException ex) {
			//pass
		}
		
		parser = new DefaultStringLiteralParser("'", '\\', IMPLICIT_INITIAL_QUOTE);
		assertEquals("foo", parser.parseStringLiteral("foo'bar'", new ParsePosition(0)));
		assertEquals("", parser.parseStringLiteral("foo'bar'", new ParsePosition(3)));
		assertEquals("bar", parser.parseStringLiteral("foo'bar'", new ParsePosition(4)));
	}

	public void testParseStringLiteral_IMPLICIT_FINAL_QUOTE() throws ParseException {
		DefaultStringLiteralParser parser = new DefaultStringLiteralParser("'", '\\', 0);

		try {
			parser.parseStringLiteral("'foo", new ParsePosition(0));
			fail("expected complaint about missing terminal quote");
		} catch (ParseException ex) {
			//pass
		}
		
		parser = new DefaultStringLiteralParser("'", '\\', IMPLICIT_FINAL_QUOTE);
		assertEquals("foo", parser.parseStringLiteral("'foo", new ParsePosition(0)));
		assertEquals("foo", parser.parseStringLiteral("'foo'", new ParsePosition(0)));
		
		ParsePosition pos = new ParsePosition(0);
		assertEquals("foo", parser.parseStringLiteral("'foo\nbar", pos));
		assertEquals(4, pos.getIndex());

		parser = new DefaultStringLiteralParser("'", '\\',IMPLICIT_FINAL_QUOTE | ALLOW_ESCAPED_LINE_BREAKS | ALLOW_LINE_BREAKS);
		assertEquals("foo\r", parser.parseStringLiteral("'foo\\\r", new ParsePosition(0)));
		assertEquals("foo\r", parser.parseStringLiteral("'foo\r", new ParsePosition(0)));

		parser = new DefaultStringLiteralParser("'", '\\',IMPLICIT_INITIAL_QUOTE |IMPLICIT_FINAL_QUOTE);
		assertEquals("", parser.parseStringLiteral("", new ParsePosition(0)));
	}

}
