package de.brightbyte.util;

import java.beans.IntrospectionException;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.brightbyte.io.IOUtil;

/**
 * Simple codec for encoding basic java data structures to strings, and decoding them again.
 * Supported are all java primitive types (resp. their wrapper classes), as well as null, 
 * Maps and Lists (may be nested arbitrarily). 
 * Caveat: cyclic strucures will cause the codec to recurse indefinitely, eventually resulting
 * in a StackOverflowError.
 */
public class StructuredDataCodec {
	
	public static interface Lookup {
		public Object get(String key);
	}
	
	public static class DefaultLookup implements Lookup {
		protected Map<String, Object> data;
		
		public DefaultLookup(Map<String, Object> data) {
			if (data==null) throw new NullPointerException();
			this.data = data;
		}

		public Object get(String key) {
			return data.get(key);
		}
	}
	
	public static interface Factory<T> {
		public T create(List<Object> params, Map<String, Object> props);
		public String getTypeName();		
		public Class<T> getTypeClass();
		public String encode(T obj);
	}
	
	public static abstract class DefaultFactory<T> implements Factory<T> {
		protected String typeName;
		protected Class<T> typeClass;
		
		public DefaultFactory(Class<T> typeClass, String typeName) {
			super();
			this.typeName = typeName;
			this.typeClass = typeClass;
		}
		
		public Class<T> getTypeClass() {
			return typeClass;
		}
		
		public String getTypeName() {
			return typeName;
		}

		public T create(List<Object> params, Map<String, Object> props) {
			try {
				return BeanUtils.createBean(typeClass, params, props);
			} catch (IllegalArgumentException e) {
				throw new IllegalArgumentException("instantiation failed", e);
			} catch (IntrospectionException e) {
				throw new IllegalArgumentException("instantiation failed", e);
			} catch (InstantiationException e) {
				throw new IllegalArgumentException("instantiation failed", e);
			} catch (IllegalAccessException e) {
				throw new IllegalArgumentException("instantiation failed", e);
			} catch (NoSuchMethodException e) {
				throw new IllegalArgumentException("instantiation failed", e);
			} catch (InvocationTargetException e) {
				throw new IllegalArgumentException("instantiation failed", e);
			}
		}

		public String encode(T obj) {
			return "new " + getTypeName() + encodeInitData(obj);
		}

		protected String encodeInitData(T obj) {
			return "()";
		}
	}
	
	protected static final Pattern defaultChunker = Pattern.compile(
			"\\s*([()\\]\\[{},;|:=]|null|true|false|[FD]?-?\\d+(\\.\\d+([eE]-?\\d+)?)?|[BSIL]?-?\\d+|\"(\\\\.|[^\"\\\\]+)*\"|'([^'\\\\]|\\\\'|\\\\\\\\)'|\\$?[\\w.]+)\\s*", Pattern.DOTALL);
	
	protected static final Pattern defaultIdentPattern = Pattern.compile("\\w+[\\w\\d]*(\\.\\w+[\\w\\d]*)*");
	
	protected static final Pattern specialCharPattern = Pattern.compile("([\\\\\"])");
	
	public static final String EOF = new String("eof");

	public static final StructuredDataCodec instance = new StructuredDataCodec();
	
	protected Pattern chunker;
	protected Pattern identPattern;
	protected Lookup lookup;
	protected Map<String, Factory> factories;
	
	public StructuredDataCodec() {
		this(null, null, null);
	}
	
	public StructuredDataCodec(Lookup lookup, Pattern chunker, Pattern identPattern) {
		if (chunker==null) chunker = defaultChunker;
		if (identPattern==null) identPattern = defaultIdentPattern;
		
		this.lookup = lookup;
		this.chunker = chunker;
		this.identPattern = identPattern;
	}
	
	public void setLookup(Lookup lookup) {
		this.lookup = lookup;
	}
	
	public void addFactory(Factory factory) {
		if (this.factories==null) this.factories = new  HashMap<String, Factory>();
		this.factories.put(factory.getTypeName(), factory);
	}
	
	protected String encodeMap(Map<?, ?> m) {
		if (m==null) return encodeValue(null);
		
		StringBuilder s = new StringBuilder();
		
		s.append('{');
		boolean first = true;
		for (Map.Entry<?, ?> e: m.entrySet()) {
			String k = encodeValue(e.getKey());
			String v = encodeValue(e.getValue());
			
			if (first) first = false;
			else s.append(',');
			
			s.append(k);
			s.append(':');
			s.append(v);
		}
		s.append('}');
		
		return s.toString();
	}
	
	protected String encodeList(List<?> l) {
		if (l==null) return encodeValue(null);
		
		StringBuilder s = new StringBuilder();
		
		s.append('[');
		boolean first = true;
		for (Object e: l) {
			String v = encodeValue(e);
			
			if (first) first = false;
			else s.append(',');
			
			s.append(v);
		}
		s.append(']');
		
		return s.toString();
	}
	
	public String encodeValue(Object obj) {
		if (obj==null) return "null";
		else if (obj instanceof Byte) return "B"+obj.toString();
		else if (obj instanceof Short) return "S"+obj.toString();
		else if (obj instanceof Integer) return "I"+obj.toString();
		else if (obj instanceof Long) return "L"+obj.toString();
		else if (obj instanceof Float) return "F"+obj.toString();
		else if (obj instanceof Double) return "D"+obj.toString();
		else if (obj instanceof Character) return "'"+((((Character)obj).charValue()=='\'')?"\\'":obj)+"'";
		else if (obj instanceof Boolean) return (((Boolean)obj).booleanValue())?"true":"false";
		else if (obj instanceof String) return quote((String)obj);
		else if (obj instanceof Map) return encodeMap((Map)obj);
		else if (obj instanceof Class) return encodeClass((Class)obj);
		else if (obj instanceof List) return encodeList((List)obj);
		else if (obj instanceof Object[]) return encodeList(Arrays.asList((Object[])obj));
		else if (factories!=null) {
			for (Factory f: factories.values()) {
				if (f.getTypeClass().isInstance(obj)) {
					return f.encode(obj);
				}
			}
		}
		
		throw new IllegalArgumentException("unsupported data type "+obj.getClass().getName());
	}
	
	protected String encodeClass(Class cls) {
		return "class " + cls.getName();
	}

	public static class Scanner {
		protected Matcher matcher;
		protected Matcher identMatcher;
		protected String text;
		protected String last;
		protected boolean matched;
		
		public Scanner(Pattern chunker, Pattern identPattern, String s) {
			text = s;
			matcher = chunker.matcher(s);
			identMatcher = identPattern.matcher(s);
			seek(0);
		}
		
		public void seek(int pos) {
			matcher.region(pos, text.length());
			matched = false;
		}
		
		public void pushback() {
			seek(matcher.regionStart()); 
		}
		
		public String next() {
			matcher.region(getPos(), text.length());
			matched = false;

			if (matcher.regionStart()>=text.length()) return EOF;
			
			if (!matcher.lookingAt()) {
				throw new IllegalArgumentException("malformed input: " + this);
			}
			else {
				matched = true;
				return matcher.group(1);
			}
		}
		
		@Override
		public String toString() {
			return getClass().getName()+"(@"+getPos()+")";
		}
		
		public int getPos() {
			if (matched) return matcher.end();
			else return matcher.regionStart();
		}
		
		public int scanValue() {
			int pos = getPos();
			
			if (pos>=text.length()) return pos;
			String chunk = next();
			scanChunk(chunk, true);
			return pos;
		}
		
		protected Object scanChunk(String chunk, boolean allowStruct) {
			if (chunk == EOF) return null;			
			char ch = chunk.charAt(0);

			if (ch=='{') {
				if (!allowStruct) throw new IllegalArgumentException("structure not allowed here!");
				return scanMap(null, "}");
			}
			else if (ch=='[') {
				if (!allowStruct) throw new IllegalArgumentException("structure not allowed here!");
				return scanList(null, "]");
			}
			else if (ch=='(') {
				if (!allowStruct) throw new IllegalArgumentException("structure not allowed here!");
				return scanList(null, ")");
			}
			else if (ch=='n' && chunk.equals("new")) {
				if (!allowStruct) throw new IllegalArgumentException("structure not allowed here!");
				return scanNew();
			}
			else if (ch=='c' && chunk.equals("class")) {
				if (!allowStruct) throw new IllegalArgumentException("structure not allowed here!");
				return scanClass();
			}
			else {
				return handleScalar(chunk);
			}
		}
		
		protected Class scanClass() {
			String type = next();
			if (type==EOF) throw new IllegalArgumentException("premature end of input, identifier expected");
			
			identMatcher.reset(type);
			if (!identMatcher.matches())   throw new IllegalArgumentException("identifier expected");
			
			return getClassByName(type);
		}
		
		protected Class getClassByName(String name) {
			return null;
		}

		protected Object scanNew() {
			String type = next();
			if (type==EOF) throw new IllegalArgumentException("premature end of input, identifier expected");
			
			identMatcher.reset(type);
			if (!identMatcher.matches())   throw new IllegalArgumentException("identifier expected");
			
			List<Object> params = null;
			Map<Object, Object> properties = null;
			
			String chunk = next();
			
			if (chunk != EOF) {
				char ch = chunk.charAt(0);
				
				if (ch=='{') {
					properties = scanMap(new HashMap<Object, Object>(), "}");
				}
				else if (ch=='[') {
					params =  scanList(new ArrayList<Object>(), "]");
				}
				else if (ch=='(') {
					params =  scanList(new ArrayList<Object>(), ")");
				} else {
					pushback();
				}
			}
			
			return instantiate(type, params, (Map<String, Object>)(Object)properties);
		}
		
		protected Object instantiate(String type, List<Object> param, Map<String, Object> props) {
			return null;
		}
		
		protected Map<Object, Object> scanMap(Map<Object, Object> into, String terminator) {
			boolean first = true;
			
			while (true) {
				String chunk = next();
				if (chunk.equals(terminator)) break;
				if (chunk==EOF) throw new IllegalArgumentException("premature end of input");
				
				if (first) first = false;
				else {
					if (!chunk.equals(",") && !chunk.equals(";") && !chunk.equals("|")) throw new IllegalArgumentException("bad input, ´,´, ´;´ or ´|´ expected");
					chunk = next();
					if (chunk==EOF) throw new IllegalArgumentException("premature end of input");
				}

				Object k = scanChunk(chunk, false);

				chunk = next();
				if (chunk==EOF) throw new IllegalArgumentException("premature end of input");
				if (!chunk.equals(":") && !chunk.equals("=")) throw new IllegalArgumentException("bad input, ´:´ or ´=´ expected");

				chunk = next();
				if (chunk==EOF) throw new IllegalArgumentException("premature end of input");
				Object v = scanChunk(chunk, true);
				
				if (into!=null) into.put(k, v);
			}
				
			return into;
		}
		
		protected List<Object> scanList(List<Object> into, String terminator) {
			boolean first = true;
			
			while (true) {
				String chunk = next();
				if (chunk.equals(terminator)) break;
				if (chunk==EOF) throw new IllegalArgumentException("premature end of input");
				
				if (first) first = false;
				else {
					if (!chunk.equals(",") && !chunk.equals(";") && !chunk.equals("|")) throw new IllegalArgumentException("bad input, ´,´, ´;´ or ´|´ expected");
					chunk = next();
					if (chunk==EOF) throw new IllegalArgumentException("premature end of input");
				}
					
				Object v = scanChunk(chunk, true);
				if (into!=null) into.add(v);
			}
				
			return into;
		}
		
		protected Object handleScalar(String chunk) {
			return null; //noop
		}
		
	}
	
	protected class Builder extends Scanner {
		protected Builder(Pattern chunker, Pattern identPatter, String s) {
			super(chunker, identPatter, s);
		}
		
		public Object buildValue() {
			String chunk = next();
			return scanChunk(chunk, true);
		}

		@Override
		protected Object handleScalar(String chunk) {
			char ch = chunk.charAt(0);
			Object v;
			
			if (ch=='n' && chunk.equals("null")) v = null;
			else if (ch=='t' && chunk.equals("true")) v = Boolean.TRUE;
			else if (ch=='f' && chunk.equals("false")) v = Boolean.FALSE;
			else if (ch=='"') v = chunk.substring(1, chunk.length()-1).replaceAll("\\\\(.)", "$1");
			else if (ch=='\'') v = chunk.equals("'\\''") ? '\'' : chunk.charAt(1);
			else if (ch=='B') v = Byte.parseByte(chunk.substring(1));
			else if (ch=='S') v = Short.parseShort(chunk.substring(1));
			else if (ch=='I') v = Integer.parseInt(chunk.substring(1));
			else if (ch=='L') v = Long.parseLong(chunk.substring(1));
			else if (ch=='F') v = Float.parseFloat(chunk.substring(1));
			else if (ch=='D') v = Double.parseDouble(chunk.substring(1));
			else if (ch=='$') {
				if (lookup==null) throw new IllegalArgumentException("No lookup provided, references not supported! Found "+chunk);
				else v = lookup.get(chunk.substring(1));
			}
			else if ((ch>='0' && ch<='9') || ch=='-') {
				if (chunk.indexOf('.')>0) v = Double.parseDouble(chunk);
				else v = Integer.parseInt(chunk);
			}
			else if (identPattern.matcher(chunk).matches()) v = chunk;  
			else throw new IllegalArgumentException("illegal input (chunk: "+chunk+"): "+this);
			
			return v;
		}

		@Override
		protected List<Object> scanList(List<Object> into, String end) {
			if (into==null) into = new ArrayList<Object>();
			super.scanList(into, end);
			return into;
		}

		@Override
		protected Map<Object, Object> scanMap(Map<Object, Object> into, String terminator) {
			if (into==null) into = new HashMap<Object, Object>();
			super.scanMap(into, terminator);
			return into;
		}
				
		protected Class getClassByName(String name) {
			try {
				return Class.forName(name, true, getClass().getClassLoader());
			} catch (ClassNotFoundException e) {
				throw new IllegalArgumentException("class not found", e);
			}
		}
		
		protected Object instantiate(String type, List<Object> params, Map<String, Object> props) {
			Factory factory = factories==null ? null : factories.get(type);
			if (factory!=null) return factory.create(params, props);
			
			try {
				return BeanUtils.createBean(type, Object.class, params, props);
			} catch (IllegalArgumentException e) {
				throw new IllegalArgumentException("instantiation failed", e);
			} catch (IntrospectionException e) {
				throw new IllegalArgumentException("instantiation failed", e);
			} catch (InstantiationException e) {
				throw new IllegalArgumentException("instantiation failed", e);
			} catch (IllegalAccessException e) {
				throw new IllegalArgumentException("instantiation failed", e);
			} catch (NoSuchMethodException e) {
				throw new IllegalArgumentException("instantiation failed", e);
			} catch (InvocationTargetException e) {
				throw new IllegalArgumentException("instantiation failed", e);
			} catch (ClassNotFoundException e) {
				throw new IllegalArgumentException("instantiation failed", e);
			}
		}
	}
	
	public Scanner scanner(String s) {
		return new Scanner(chunker, identPattern, s);
	}
	
	protected Builder builder(String s) {
		return new Builder(chunker, identPattern, s);
	}
	
	public Object decodeValue(String s) {
		Builder builder = builder(s);
		return builder.buildValue();
		
		/*
		Scanner scanner = scanner(s);
		Object obj = decodeValue(scanner);
		if (obj == EOF) throw new IllegalArgumentException("premature end of input");
		return obj;
		*/
	}

	
	public String quote(String s) {
		return "\"" + specialCharPattern.matcher(s).replaceAll("\\\\$1") + "\"";
	}

	public Map<String, Object> decodeValues(Properties props) {
		Map<String, Object> m = new HashMap<String, Object>();
		Enumeration names = props.propertyNames();
		while (names.hasMoreElements()) {
			String n = (String)names.nextElement();
			String s = props.getProperty(n);
			Object v = decodeValue(s);
			
			m.put(n, v);
		}
		
		return m;
	}
	
	public Map<String, Object> decodeValues(Map<String, String> props) {
		Map<String, Object> m = new HashMap<String, Object>();
		for (Map.Entry<String, String> e: props.entrySet()) {
			String n = e.getKey();
			String s = e.getValue();
			Object v = decodeValue(s);
			
			m.put(n, v);
		}
		
		return m;
	}

	public Map<Object, Object> decodeMap(String s) {
		Scanner builder = builder(s);
		return builder.scanMap(null, EOF);
	}

	public Map<Object, Object> loadMap(Reader rd) throws IOException {
		return decodeMap( IOUtil.slurp(rd) );
	}

	public Map<Object, Object> loadMap(InputStream in, String enc) throws IOException {
		return decodeMap( IOUtil.slurp(in, enc) );
	}

	public Map<Object, Object> loadMap(URL u) throws IOException {
		return decodeMap( IOUtil.slurp(u) );
	}

	public Map<Object, Object> loadMap(File f, String enc) throws IOException {
		return decodeMap( IOUtil.slurp(f, enc) );
	}

	public Map<Object, Object> loadMap(URL f, String enc) throws IOException {
		return decodeMap( IOUtil.slurp(f, enc) );
	}
}
