package de.brightbyte.util;

import java.text.ParsePosition;
import java.util.Arrays;
import java.util.regex.MatchResult;

import de.brightbyte.io.Output;

/**
 * Matcher for finding the location of any one of a number of substrings in a string.
 * Does not support backtracking partial matches, that is, when looking for "ababc" 
 * in "abababc", the search will fail.
 * @author daniel
 */
public class SubstringMatcher implements MatchResult {

	private ParsePosition pos;
	private CharSequence text;
	private String[] patterns;
	private int[] offsets;
	
	private int matched = -1;
	private int start;
	private int end;

	private Output trace;
	
	public SubstringMatcher(String... p) {
		this(new ParsePosition(0), p);
	}
	
	public SubstringMatcher(ParsePosition pos, String... p) {
		if (pos==null) throw new NullPointerException();
		this.pos = pos;
		
		patterns = p;
		offsets = new int[p.length];
	}
	
	public void reset() {
		reset(text);
	}
	
	public void reset(CharSequence t) {
		pos.setIndex( regionStart() );
		text = t;
		matched = -1;
	}

	public void reset(CharSequence t, int idx) {
		reset(t);
		setIndex(idx);
	}
	
	public boolean find(int idx) {
		pos.setIndex(idx);
		return find();
	}

	public boolean find() {
		int i = pos.getIndex();
		
		matched = -1;
		start = -1;
		end = -1;
		
		if (patterns.length==0) { //short cirquit
			pos.setIndex(regionEnd());
			if (trace!=null) trace.println("fail (no patterns)");
			return false;
		}
		
		int c = regionEnd();

		if (i>=c) { //short cirquit
			pos.setIndex(i);
			if (trace!=null) trace.println("fail (nothing to match)");
			return false;
		}
		
		Arrays.fill(offsets, 0);
		
		while (i<c) {
			char ch = text.charAt(i);
			
			for (int j=0; j<offsets.length; j++) {
				char chj = patterns[j].charAt(offsets[j]);
				
				if (ch!=chj) {
					offsets[j] = 0;
					chj = patterns[j].charAt(offsets[j]);
				}
				
				if (ch==chj) {
					offsets[j]++;
					if (offsets[j]>=patterns[j].length()) {
						i++;
						end = i;
						start = end - patterns[j].length(); 
						matched = j;
						pos.setIndex(i);
						
						if (trace!=null) {
							trace.println("found " + patterns[j] + " @" + start);
						}
						
						return true;
					}
				}
			}
			
			i++;
		}
		
		pos.setIndex(i);

		if (trace!=null) trace.println("fail (no match found)");
		return false;
	}
	
	public int matchedGroup() {
		return matched+1;
	}

	public int regionStart() {
		return 0;
	}

	public int regionEnd() {
		return text.length();
	}

	public int end() {
		return end;
	}

	public int end(int group) {
		if (group>0 && group-1 != matched) return -1;
		return end;
	}

	public String group() {
		return matched < 0 ? null : patterns[matched];
	}

	public String group(int group) {
		if (group==0) return matched < 0 ? null : patterns[matched];
		else if (group-1==matched) return patterns[matched];
		else return null;
	}

	public int groupCount() {
		return patterns.length;
	}

	public int start() {
		return start;
	}

	public int start(int group) {
		if (group>0 && group-1 != matched) return -1;
		return start;
	}
	
	public void setIndex(int idx) {
		pos.setIndex(idx);
	}
	
	public int index() {
		return pos.getIndex();
	}

	public CharSequence getText() {
		return text;
	}
	
	public void setTrace(Output trace) {
		this.trace = trace;
	}
	
}
