package de.brightbyte.xml;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.brightbyte.util.CollectionUtils;
import de.brightbyte.util.StringUtils;

public class XmlBuilder {
	
	protected Appendable target = new StringBuilder();
	
	public XmlBuilder() {
		this(new StringBuilder());
	}
		
	public XmlBuilder(Appendable target) {
		this.target = target;
	}
		
	public static String escape(String s) {
		return XmlUtil.escape(s);
	}
	
	@Override
	public String toString() {
		return getTarget().toString();
	}
	
	public Appendable getTarget() {
		return target;
	}
	
	public XmlBuilder append(CharSequence html) throws IOException {
		target.append(html);
		return this;
	}

	public XmlBuilder append(char ch) throws IOException {
		target.append(ch);
		return this;
	}

	public XmlBuilder appendEscaped(String s) throws IOException {
		append(escape(s));
		return this;
	}

	public XmlBuilder appendProlog(String version, String encoding) throws IOException {
		append("<?xml");

		if (version!=null) {
			append(" version=\"");
			append(version);
			append("\"");
		}
		
		if (encoding!=null) {
			append(" encoding=\"");
			append(encoding);
			append("\"");
		}
		
		append("?>");
		endProcessingInstruction();
		return this;
	}
	
	public XmlBuilder appendDoctype(String name, String sys, String pub) throws IOException {
		append("<!DOCTYPE ");
		append(name);
		append(' ');

		if (pub==null) {
			append("PUBLIC ");
			append('"');
			append(sys);
			append('"');
			append(' ');
		}
		else {
			append("SYSTEM ");
		}

		append('"');
		append(sys);
		append('"');
		
		return this;
	}
	
	public XmlBuilder startProcessingInstruction(String target) throws IOException {
		append("<?");
		append(target);
		append(' ');
		return this;
	}
	
	public XmlBuilder endProcessingInstruction() throws IOException {
		append("?>");
		return this;
	}
	
	public XmlBuilder appendProcessingInstruction(String target, String content) throws IOException {
		startProcessingInstruction(target);
		append(content);
		endProcessingInstruction();
		return this;
	}
	
	public XmlBuilder startComment(String target) throws IOException {
		append("<!--");
		append(target);
		append(' ');
		return this;
	}
	
	public XmlBuilder endComment() throws IOException {
		append("-->");
		return this;
	}
	
	public XmlBuilder appendComment(String target, String content) throws IOException {
		content.replaceAll("--", "- -");
		
		startProcessingInstruction(target);
		append(content);
		endProcessingInstruction();
		return this;
	}
	
	public XmlBuilder startTag(String t, Map<String, ? extends Object>... attrs) throws IOException {
		openTag(t, attrs);
		closeTag(false);
		return this;
	}
	
	public XmlBuilder appendEscapedTag(String t, String s, Map<String, ? extends Object>... attrs) throws IOException {
		appendTag(t, escape(s), attrs);
		return this;
	}
	
	public XmlBuilder appendTag(String t, CharSequence html, Map<String, ? extends Object>... attrs) throws IOException {
		openTag(t, attrs);
		if (html==null) {
			closeTag(true);
		}
		else {
			closeTag(false);
			append(html);
			endTag(t);
		}
		return this;
	}
	
	public XmlBuilder openTag(String t, Map<String, ? extends Object>... attrs) throws IOException {
		append("<");
		append(t);
		appendAttributes(attrs);
		return this;
	}
	
	public XmlBuilder closeTag(boolean end) throws IOException {
		if (end) append("/");
		append(">");
		return this;
	}

	public XmlBuilder endTag(String t) throws IOException {
		append("</");
		append(t);
		append(">");
		return this;
	}

	public XmlBuilder appendAttributes(Map<String, ? extends Object>... attrs) throws IOException {
		if (attrs==null || attrs.length==0 || attrs[0]==null) return this;
		
		Map<String, ? extends Object> m = new HashMap<String, Object>();
		
		for (Map<String, ? extends Object> attr: attrs) {
			CollectionUtils.deepMerge((Map<String, Object>)m, (Map<String, Object>)attr);
		}
		
		attributesInternal(m);
		return this;
	}

	public XmlBuilder appendAttribute(String prefix, String name, Object value) throws IOException {
		if (name==null || value==null || value.equals(Boolean.FALSE)) return this;
		
		String s = attributeValue(name, value);
		
		append(' ');
		
		if (prefix!=null) {
			append(prefix);
			append(':');
		}
		
		append(name);
		append('=');
		append('"');
		append(XmlUtil.escape(s));
		append('"');
		return this;
	}

	public XmlBuilder appendAttribute(String name, Object value) throws IOException {
		return appendAttribute(null, name, value);
	}
	
	private void attributesInternal(Map<String, ? extends Object> attr) throws IOException {
		if (attr==null || attr.isEmpty());
		
		for (Map.Entry<String, ? extends Object> e: attr.entrySet()) {
			String n = e.getKey();
			Object v = e.getValue();
			
			appendAttribute(n, v);
		}
	}

	public static String attributeValue(String n, Object v) {
		if (v.equals(Boolean.TRUE)) v = n; 
		
		String s;
		if (v instanceof Map) {
			if (n.equals("style")) {
				s = styles((Map<String, String>)v);
			}
			else {
				throw new IllegalArgumentException("value for attribute "+n+" must not be a map");
			}
		}
		else if (v instanceof List) {
			if (n.equals("id")) throw new IllegalArgumentException("id attribute can not be a list");
			s = StringUtils.join(" ", (List)v);
		}
		else if (v instanceof Object[]) {
			if (n.equals("id")) throw new IllegalArgumentException("id attribute can not be a list");
			s = StringUtils.join(" ", (Object[])v);
		}
		else if (v instanceof String || v.getClass().isPrimitive()) {
			s = v.toString();
		}
		else throw new IllegalArgumentException("bad type for value of attribute "+n+": "+v.getClass());
		
		return s;
	}

	public static String styles(Map<String, String> st) {
		if (st==null || st.isEmpty()) return "";
		
		StringBuilder s = new StringBuilder();
		
		for (Map.Entry<String, String> e: st.entrySet()) {
			String n = e.getKey();
			String v = e.getValue();
			
			if (v==null) continue;
			
			s.append(' ');
			s.append(n);
			s.append(':');
			s.append(v);
			s.append(';');
		}
		
		return s.toString();
	}
}
